@@ -1,20 +1,20 @@
 package Graph;
 
-BEGIN {
-    if (0) { # SET TO ZERO FOR TESTING AND RELEASES!
-	$SIG{__DIE__ } = \&CarpConfess;
-	$SIG{__WARN__} = \&CarpConfess;
-    }
-    sub CarpConfess () { require Carp; Carp::confess }
-}
-
 use strict;
 
+BEGIN {
+    if (0) { # SET THIS TO ZERO FOR TESTING AND RELEASES!
+	$SIG{__DIE__ } = \&__carp_confess;
+	$SIG{__WARN__} = \&__carp_confess;
+    }
+    sub __carp_confess { require Carp; Carp::confess(@_) }
+}
+
 use Graph::AdjacencyMap qw(:flags :fields);
 
 use vars qw($VERSION);
 
-$VERSION = '0.68';
+$VERSION = '0.70';
 
 require 5.006; # Weak references are required.
 
@@ -59,19 +59,19 @@ sub _U () { 5 } # Union-Find.
 # - _UNORDUNIQ: just a union of _UNORD and UNIQ
 # Vertices are assumed to be _UNORDUNIQ; edges assume none of these flags.
 
-use Graph::Attribute array => _A, map => "graph";
+use Graph::Attribute array => _A, map => 'graph';
 
 sub _COMPAT02 () { 0x00000001 }
 
 sub stringify {
     my $g = shift;
     my $o = $g->is_undirected;
-    my $e = $o ? "=" : "-";
+    my $e = $o ? '=' : '-';
     my @e =
 	map {
 	    my @v =
 		map {
-		    ref($_) eq 'ARRAY' ? "[" . join(" ", @$_). "]" : "$_"
+		    ref($_) eq 'ARRAY' ? "[" . join(" ", @$_) . "]" : "$_"
 		}
 	    @$_;
 	    join($e, $o ? sort { "$a" cmp "$b" } @v : @v) } $g->edges05;
@@ -749,11 +749,13 @@ sub _edges_from {
 	if (wantarray) {
 	    if ($o) {
 		while (my ($ei, $ev) = each %{ $Ei }) {
+		    next unless @$ev;
 		    push @e, [ $ei, $ev ]
 			if ($ev->[0] == $vi || $ev->[-1] == $vi) && !$ev{$ei}++;
 		}
 	    } else {
 		while (my ($ei, $ev) = each %{ $Ei }) {
+		    next unless @$ev;
 		    push @e, [ $ei, $ev ]
 			if $ev->[0] == $vi && !$ev{$ei}++;
 		}
@@ -761,11 +763,12 @@ sub _edges_from {
 	} else {
 	    if ($o) {
 		while (my ($ei, $ev) = each %{ $Ei }) {
-		    $en++
-			if ($ev->[0] == $vi || $ev->[-1] == $vi);
+		    next unless @$ev;
+		    $en++ if ($ev->[0] == $vi || $ev->[-1] == $vi);
 		}
 	    } else {
 		while (my ($ei, $ev) = each %{ $Ei }) {
+		    next unless @$ev;
 		    $en++ if $ev->[0] == $vi;
 		}
 	    }
@@ -797,11 +800,13 @@ sub _edges_to {
 	if (wantarray) {
 	    if ($o) {
 		while (my ($ei, $ev) = each %{ $Ei }) {
+		    next unless @$ev;
 		    push @e, [ $ei, $ev ]
 			if ($ev->[-1] == $vi || $ev->[0] == $vi) && !$ev{$ei}++;
 		}
 	    } else {
 		while (my ($ei, $ev) = each %{ $Ei }) {
+		    next unless @$ev;
 		    push @e, [ $ei, $ev ]
 			if $ev->[-1] == $vi && !$ev{$ei}++;
 		}
@@ -809,10 +814,12 @@ sub _edges_to {
 	} else {
 	    if ($o) {
 		while (my ($ei, $ev) = each %{ $Ei }) {
+		    next unless @$ev;
 		    $en++ if $ev->[-1] == $vi || $ev->[0] == $vi;
 		}
 	    } else {
 		while (my ($ei, $ev) = each %{ $Ei }) {
+		    next unless @$ev;
 		    $en++ if $ev->[-1] == $vi;
 		}
 	    }
@@ -1158,7 +1165,7 @@ sub delete_cycle {
 
 sub has_cycle {
     my $g = shift;
-    $g->has_path(@_, $_[0]);
+    @_ ? $g->has_path(@_, $_[0]) : 0;
 }
 
 sub has_a_cycle {
@@ -2009,7 +2016,7 @@ sub _get_options {
 # Random constructors and accessors.
 #
 
-sub fisher_yates_shuffle (@) {
+sub __fisher_yates_shuffle (@) {
     # From perlfaq4, but modified to be non-modifying.
     my @a = @_;
     my $i = @a;
@@ -2030,7 +2037,7 @@ BEGIN {
     # bleadperl that calls itself 5.9.3 but doesn't yet have the
     # patches, oh, well.
     *_shuffle = $^P && $] < 5.009003 ?
-	\&fisher_yates_shuffle : \&List::Util::shuffle;
+	\&__fisher_yates_shuffle : \&List::Util::shuffle;
 }
 
 sub random_graph {
@@ -2532,13 +2539,14 @@ sub _connected_components {
 		@cci{ keys %icci } = values %icci;
 	    }
 	} else {
-	    my %r; @r{ $g->unique_vertices } = ();
+	    my @u = $g->unique_vertices;
+	    my %r; @r{ @u } = @u;
 	    my $froot = sub {
-		wantarray ? ( keys %r ? scalar each %r: ()) : each %r;
+		(each %r)[1];
 	    };
 	    my $nroot = sub {
 		$cc++;
-		wantarray ? ( keys %r ? scalar each %r: ()) : each %r;
+		(each %r)[1];
 	    };
 	    my $t = Graph::Traversal::DFS->new($g,
 					       first_root => $froot,
@@ -2607,17 +2615,22 @@ sub same_connected_components {
     }
 }
 
-my $super_component = sub { join("+", sort @{ $_[0] }) };
+my $super_component = sub { join("+", sort @_) };
 
 sub connected_graph {
-    my $g = shift;
+    my ($g, %opt) = @_;
     $g->expect_undirected;
     my $cg = Graph->new(undirected => 1);
     if ($g->has_union_find && $g->vertices == 1) {
 	$cg->add_vertices($g->vertices);
     } else {
+	my $sc_cb =
+	    exists $opt{super_component} ?
+		$opt{super_component} : $super_component;
 	for my $cc ( $g->connected_components() ) {
-	    $cg->add_vertex($super_component->($cc));
+	    my $sc = $sc_cb->(@$cc);
+	    $cg->add_vertex($sc);
+	    $cg->set_vertex_attribute($sc, 'subvertices', [ @$cc ]);
 	}
     }
     return $cg;
@@ -2820,18 +2833,28 @@ sub strongly_connected_graph {
 
     $u->dfs;
 
-    my $hypervertex = $super_component;
+    my $sc_cb;
+    my $hv_cb;
 
-    _opt_get(\%attr, hypervertex => \$hypervertex);
+    _opt_get(\%attr, super_component => \$sc_cb);
+    _opt_get(\%attr, hypervertex => \$hv_cb);
     _opt_unknown(\%attr);
 
+    if (defined $hv_cb && !defined $sc_cb) {
+	$sc_cb = sub { $hv_cb->( [ @_ ] ) };
+    }
+    unless (defined $sc_cb) {
+	$sc_cb = $super_component;
+    }
+
     my $s = Graph->new;
 
     my %c;
     my @s;
     for (my $i = 0; $i <  @c; $i++) {
 	my $c = $c[$i];
-	$s->add_vertex( $s[$i] = $hypervertex->($c) );
+	$s->add_vertex( $s[$i] = $sc_cb->(@$c) );
+	$s->set_vertex_attribute($s[$i], 'subvertices', [ @$c ]);
 	for my $v (@$c) {
 	    $c{$v} = $i;
 	}
@@ -2904,84 +2927,102 @@ sub biconnectivity {
 	my %L = ( $r => 1 );
 	my @S = ( $r );
 	my %A;
-	for my $v ($g->vertices) {
-	    @{ $A{ $v } }{ $g->successors( $v ) } = ();
+	my @V = $g->vertices;
+
+	for my $v (@V) {
+	    my @s = $g->successors( $v );
+	    @{ $A{ $v } }{ @s } = @s;
 	}
-	my @C;
-	my $Avok;
-	do {
-	    my $w;
-	    do {
-		$w = first { !$U{ $v }{ $_ } } _shuffle keys %{ $A{ $v } };
-		if (defined $w) {
-		    $U{ $v }{ $w }++;
-		    $U{ $w }{ $v }++;
-		    if ($I{ $w } == 0) {
-			$P{ $w } = $v;
-			$i++;
-			$I{ $w } = $i;
-			$L{ $w } = $i;
-			push @S, $w;
-			$v = $w;
-		    } else {
-			$L{ $v } = $I{ $w } if $I{ $w } < $L{ $v };
-		    }
-		}
-	    } while (defined $w);
-	    if (!defined $P{ $v }) {
-		# Do nothing.
-	    } elsif ($P{ $v } ne $r) {
-		if ($L{ $v } < $I{ $P{ $v } }) {
-		    $L{ $P{ $v } } = $L{ $v } if $L{ $v } < $L{ $P{ $v } };
-		} else {
-		    $AP{ $P{ $v } }++;
-		    push @C, _make_bcc(\@S, $v, $P{ $v } );
-		}
-	    } else {
-		my $e;
-		for my $w (_shuffle keys %{ $A{ $r } }) {
-		    unless ($U{ $r }{ $w }) {
-			$e = $r;
-			last;
-		    }
-		}
-		$AP{ $e }++ if defined $e;
-		push @C, _make_bcc(\@S, $v, $r);
-	    }
-	    $v = defined $P{ $v } ? $P{ $v } : $r;
-	    $Avok = 0;
-	    if (defined $v) {
-		$Avok = 0;
-		if (keys %{ $A{ $v } }) {
-		    if (!exists $P{ $v }) {
-			for my $w (keys %{ $A{ $v } }) {
-			    $Avok++ if $U{ $v }{ $w };
-			}
-			$Avok = 0 unless $Avok == keys %{ $A{ $v } };
-		    }
-		} else {
-		    $Avok = 1;
-		}
-	    }
-	} until ($Avok);
 
 	my %V2BC;
-	for (my $i = 0; $i < @C; $i++) {
-	    for my $v (@{ $C[ $i ]}) {
-		$V2BC{ $v }{ $i }++;
+	my @BR;
+	my @BC;
+
+	my @C;
+	my $Avok;
+
+	my %T; @T{ @V } = @V;
+
+	while (keys %T) {
+	    do {
+		my $w;
+		do {
+		    $w = first { !$U{ $v }{ $_ } } _shuffle values %{ $A{ $v } };
+		    if (defined $w) {
+			$U{ $v }{ $w }++;
+			$U{ $w }{ $v }++;
+			if ($I{ $w } == 0) {
+			    $P{ $w } = $v;
+			    $i++;
+			    $I{ $w } = $i;
+			    $L{ $w } = $i;
+			    push @S, $w;
+			    $v = $w;
+			} else {
+			    $L{ $v } = $I{ $w } if $I{ $w } < $L{ $v };
+			}
+		    }
+		} while (defined $w);
+		if (!defined $P{ $v }) {
+		    # Do nothing.
+		} elsif ($P{ $v } ne $r) {
+		    if ($L{ $v } < $I{ $P{ $v } }) {
+			$L{ $P{ $v } } = $L{ $v } if $L{ $v } < $L{ $P{ $v } };
+		    } else {
+			$AP{ $P{ $v } } = $P{ $v };
+			push @C, _make_bcc(\@S, $v, $P{ $v } );
+		    }
+		} else {
+		    my $e;
+		    for my $w (_shuffle keys %{ $A{ $r } }) {
+			unless ($U{ $r }{ $w }) {
+			    $e = $r;
+			    last;
+			}
+		    }
+		    $AP{ $e } = $e if defined $e;
+		    push @C, _make_bcc(\@S, $v, $r);
+		}
+		$v = defined $P{ $v } ? $P{ $v } : $r;
+		$Avok = 0;
+		if (defined $v) {
+		    $Avok = 0;
+		    if (keys %{ $A{ $v } }) {
+			if (!exists $P{ $v }) {
+			    for my $w (keys %{ $A{ $v } }) {
+				$Avok++ if $U{ $v }{ $w };
+			    }
+			    $Avok = 0 unless $Avok == keys %{ $A{ $v } };
+			}
+		    } else {
+			$Avok = 1;
+		    }
+		}
+	    } until ($Avok);
+
+	    last unless @C;
+
+	    for (my $i = 0; $i < @C; $i++) {
+		for my $v (@{ $C[ $i ]}) {
+		    $V2BC{ $v }{ $i }++;
+		    delete $T{ $v };
+		}
+	    }
+
+	    for (my $i = 0; $i < @C; $i++) {
+		if (@{ $C[ $i ] } == 2) {
+		    push @BR, $C[ $i ];
+		} else {
+		    push @BC, $C[ $i ];
+		}
+	    }
+
+	    if (keys %T) {
+		$r = $v = each %T;
 	    }
 	}
 
-	my @BR;
-	my @BC;
-	for (my $i = 0; $i < @C; $i++) {
-	    if (@{ $C[ $i ] } == 2) {
-		push @BR, $C[ $i ];
-	    } else {
-		push @BC, $C[ $i ];
-	    }
-	}
-	$bcc = [ [keys %AP], \@BC, \@BR, \%V2BC ];
+	$bcc = [ [values %AP], \@BC, \@BR, \%V2BC ];
 	$g->set_graph_attribute('_bcc', [ $g->[ _G ], $bcc ]);
     }
 
@@ -3057,11 +3098,15 @@ sub same_biconnected_components {
 }
 
 sub biconnected_graph {
-    my $g = shift;
+    my ($g, %opt) = @_;
     my ($bc, $v2bc) = ($g->biconnectivity)[1, 3];
     my $bcg = Graph::Undirected->new;
+    my $sc_cb =
+	exists $opt{super_component} ?
+	    $opt{super_component} : $super_component;
     for my $c (@$bc) {
-	$bcg->add_vertex($super_component->($c));
+	$bcg->add_vertex(my $s = $sc_cb->(@$c));
+	$bcg->set_vertex_attribute($s, 'subvertices', [ @$c ]);
     }
     my %k;
     for my $i (0..$#$bc) {
@@ -3074,8 +3119,8 @@ sub biconnected_graph {
 		for my $u (@u) {
 		    if (exists $j{ $u }) {
 			unless ($k{ $i }{ $j }++) {
-			    $bcg->add_edge($super_component->($bc->[$i]),
-					   $super_component->($bc->[$j]));
+			    $bcg->add_edge($sc_cb->(@{$bc->[$i]}),
+					   $sc_cb->(@{$bc->[$j]}));
 			}
 			last;
 		    }
@@ -3098,7 +3143,7 @@ sub bridges {
 
 sub _SPT_add {
     my ($g, $h, $HF, $r, $attr, $unseen, $etc) = @_;
-    for my $s ( grep { $unseen->{ $_ } } $g->successors( $r ) ) {
+    for my $s ( grep { exists $unseen->{ $_ } } $g->successors( $r ) ) {
 	my $t = $g->get_edge_attribute( $r, $s, $attr ) || 1;
 	if ($t < 0) {
 	    require Carp;
@@ -3108,6 +3153,7 @@ sub _SPT_add {
 	    $etc->{ $s } = ($etc->{ $r } || 0) + $t;
 	    # print "$r - $s : setting $s to $etc->{ $s }\n";
 	    $h->set_vertex_attribute( $s, $attr, $etc->{ $s } );
+	    $h->set_vertex_attribute( $s, 'p', $r );
 	    $HF->add( Graph::SPTHeapElem->new($r, $s, $t) );
 	}
     }
@@ -3115,17 +3161,46 @@ sub _SPT_add {
 
 sub SPT_Dijkstra {
     my $g = shift;
-    my %etc;
-    $g->_heap_walk($g->new(), \&_SPT_add, \%etc, @_);
+    my %opt = @_ == 1 ? (first_root => $_[0]) : @_;
+    my $first_root = $opt{ first_root };
+    unless (defined $first_root) {
+	$opt{ first_root} = $first_root = $g->random_vertex();
+    }
+    my $spt_di = $g->get_graph_attribute('_spt_di');
+    unless (defined $spt_di && exists $spt_di->{ $first_root } && $spt_di->{ $first_root }->[ 0 ] == $g->[ _G ]) {
+	my %etc;
+	my $sptg = $g->_heap_walk($g->new(), \&_SPT_add, \%etc, %opt);
+	$spt_di->{ $first_root } = [ $g->[ _G ], $sptg ];
+	$g->set_graph_attribute('_spt_di', $spt_di);
+    }
+    return $spt_di->{ $first_root }->[ 1 ];
 }
 
 *SSSP_Dijkstra = \&SPT_Dijkstra;
 
 *single_source_shortest_paths = \&SPT_Dijkstra;
 
-sub SPT_Bellman_Ford {
-    my $g = shift;
-    my ($opt, $unseenh, $unseena, $r, $next, $code, $attr) = $g->_root_opt(@_);
+sub SP_Dijkstra {
+    my ($g, $u, $v) = @_;
+    my $sptg = $g->SPT_Dijkstra(first_root => $u);
+    my @path = ($v);
+    my %seen;
+    my $V = $g->vertices;
+    my $p;
+    while (defined($p = $sptg->get_vertex_attribute($v, 'p'))) {
+	last if exists $seen{$p};
+	push @path, $p;
+	$v = $p;
+	$seen{$p}++;
+	last if keys %seen == $V;
+    }
+    # @path = () if @path && $path[-1] ne $u;
+    return reverse @path;
+}
+
+sub _SPT_Bellman_Ford {
+    my ($g, $opt, $unseenh, $unseena, $r, $next, $code, $attr) = @_;
+
     my %d;
     $d{ $r } = 0;
     my %p;
@@ -3170,21 +3245,57 @@ sub SPT_Bellman_Ford {
 	}
     }
 
-    my $h = $g->new();
+    return (\%p, \%d );
+}
 
-    for my $v (keys %p) {
-	my $u = $p{ $v };
-	$h->add_edge( $u, $v );
-	$h->set_edge_attribute( $u, $v, $attr,
-				$g->get_edge_attribute($u, $v, $attr));
-	$h->set_vertex_attribute( $v, $attr, $d{ $v } );
+sub SPT_Bellman_Ford {
+    my $g = shift;
+
+    my ($opt, $unseenh, $unseena, $r, $next, $code, $attr) = $g->_root_opt(@_);
+
+    unless (defined $r) {
+	$r = $g->random_vertex();
     }
 
-    return $h;
+    my $spt_bf = $g->get_graph_attribute('_spt_bf');
+    unless (defined $spt_bf && exists $spt_bf->{ $r } && $spt_bf->{ $r }->[ 0 ] == $g->[ _G ]) {
+	my ($p, $d) = $g->_SPT_Bellman_Ford($opt, $unseenh, $unseena, $r, $next, $code, $attr);
+	my $h = $g->new();
+	for my $v (keys %$p) {
+	    my $u = $p->{ $v };
+	    $h->add_edge( $u, $v );
+	    $h->set_edge_attribute( $u, $v, $attr,
+				    $g->get_edge_attribute($u, $v, $attr));
+	    $h->set_vertex_attribute( $v, $attr, $d->{ $v } );
+	    $h->set_vertex_attribute( $v, 'p', $u );
+	}
+	$spt_bf->{ $r } = [ $g->[ _G ], $h ];
+	$g->set_graph_attribute('_spt_bf', $spt_bf);
+    }
+
+    return $spt_bf->{ $r }->[ 1 ];
 }
 
 *SSSP_Bellman_Ford = \&SPT_Bellman_Ford;
 
+sub SP_Bellman_Ford {
+    my ($g, $u, $v) = @_;
+    my $sptg = $g->SPT_Dijkstra(first_root => $u);
+    my @path = ($v);
+    my %seen;
+    my $V = $g->vertices;
+    my $p;
+    while (defined($p = $sptg->get_vertex_attribute($v, 'p'))) {
+	last if exists $seen{$p};
+	push @path, $p;
+	$v = $p;
+	$seen{$p}++;
+	last if keys %seen == $V;
+    }
+    @path = () if @path && "$path[-1]" ne "$u";
+    return reverse @path;
+}
+
 ###
 # Transitive Closure.
 #
